Require Export Classical.
Require Export ClassicalChoice.
Require Export Compare_dec.
Require Export Omega.
Require Export List.
Require Export ZArith.
Require Export Zwf.

Ltac caseEq f := generalize (refl_equal f); pattern f at -1; case f.

(* SECTION: Proving Tarski's fixpoint theorem. *)

Definition antisymetric (A:Set)(R:A->A->Prop) :=
    forall x y, R x y -> R y x -> x = y.

Implicit Arguments antisymetric.
Definition upper_bound (A:Set)(R:A->A->Prop)(x:A)(f:nat -> A) :=
  forall n, R (f n) x.

Implicit Arguments upper_bound.

Definition lub (A:Set)(R:A->A->Prop)(x:A)(f:nat -> A) :=
  upper_bound R x f /\ forall y, upper_bound R y f -> R x y.

Implicit Arguments lub.

Theorem least_unique :
  forall (A:Set)(R:A->A->Prop), antisymetric R ->
  forall (P : A -> Prop)(x y: A),
  P x /\ (forall z, P z -> R x z) ->
  P y /\ (forall z, P z -> R y z) -> x = y.
intros A R Ha P x y [HPx Hlx] [HPy Hly]; apply Ha; auto.
Qed.

Theorem unique_lub : 
  forall (A:Set)(R:A->A->Prop)(f:nat->A)(x y: A),
     antisymetric R ->
     lub R x f -> lub R y f -> x = y.
intros A R f x y Ha.
exact (least_unique _ _ Ha (fun x => upper_bound R x f) x y).
Qed.

Definition complete (A:Set)(R:A->A->Prop) :=
  forall f:nat->A,
  (forall n, R (f n) (f (S n))) ->
  exists x: A, lub R x f.

Implicit Arguments complete.

Definition monotonic 
  (A B:Set)(R:A->A->Prop)(R' : B -> B -> Prop) (f:A -> B) :=
    forall x y, R x y -> R' (f x) (f y).

Implicit Arguments monotonic.

Definition continuous
  (A B:Set)(R:A->A->Prop)(R' : B -> B -> Prop) (f:A -> B) :=
  forall (c: nat -> A),  (forall n, R (c n) (c (S n))) ->
    forall (l:A), lub R l c -> 
    lub R' (f l) (fun n => f (c n)).

Implicit Arguments continuous.

Theorem continuous_imp_monotonic :
  forall (A B:Set)(R:A->A->Prop)(R':B->B->Prop)(f:A->B),
  (forall x, R x x) ->
  continuous R R' f -> monotonic R R' f.
intros A B R R' f Hr Hc x y Hxy.
assert (Hl : lub R' (f y) (fun n => f (match n with 0 => x | _ => y end))).
apply (Hc (fun n => match n with 0 => x | _ => y end)).
destruct n; auto.
split.
unfold upper_bound; destruct n; auto.
intros b H; exact (H 1).
destruct Hl as [Hu Hl']; exact (Hu 0).
Qed.

(* This theorem is not really used, but I keep it as a way to
   show that my definition of being continuous is close to the
   conventional one. *)

Theorem continuous_reverse :
  forall (A B:Set)(R:A->A->Prop)(R':B->B->Prop)(f:A->B)
  (c :nat->A),
  antisymetric R' ->
  (forall n, R (c n)(c (S n))) ->
  continuous R R' f ->
  forall (l:A), lub R l c ->
  forall (l':B), lub R' l' (fun n => f (c n)) ->
   l' = f l.
intros A B R R' f c Ha Hch Hct l Hlub l' Hlub'.
assert (Hlub2 : lub R' (f l)(fun n => f (c n))).
exact (Hct c Hch l Hlub).
apply (unique_lub B R' (fun n => f (c n))); assumption.
Qed.

Definition least_fixpoint (A:Set)(R:A->A->Prop)(f:A->A)(x:A):Prop :=
  x = f x /\ (forall y, y = f y -> R x y).

Implicit Arguments least_fixpoint.

Fixpoint iter (A:Set)(f:A->A)(k:nat)(x:A) {struct k} : A :=
  match k with 0 => x | S p => f (iter A f p x) end.

Section Tarski_theory.

Variables (A:Set) (R : A->A->Prop).

Hypotheses (Hr : forall x:A, R x x)
   (Ha : antisymetric R) (Ht : forall x y z:A, R x y -> R y z -> R x z)
   (Hc : complete R).

Variable bot : A.

Hypothesis Hb : forall x, R bot x.

Variable f : A -> A.

Hypotheses (Hct : continuous R R f).

Theorem f_monotonic : monotonic R R f.
apply continuous_imp_monotonic; auto.
Qed.

Theorem Hchain_iter_f_bot : forall n, R (iter A f n bot) (iter A f (S n) bot).
induction n.
simpl; apply Hb.
apply (f_monotonic (iter A f n bot) (iter A f (S n) bot)); assumption.
Qed.

Theorem upper_bound_iterates_add_f :
   forall phi : A, 
       upper_bound R phi (fun n => iter A f n bot) ->
       upper_bound R phi (fun n => f (iter A f n bot)).
intros phi Hup p; apply (Hup (S p)).
Qed.

Theorem f_upper_bound_iterates :
   forall phi : A, 
       upper_bound R phi (fun n => iter A f n bot) ->
       upper_bound R (f phi) (fun n => iter A f n bot).
intros phi Hup [ | p].
apply Hb.
apply (f_monotonic (iter A f p bot)); apply (Hup p).
Qed.

Theorem lub_iterates_fixpoint : 
   forall phi : A, lub R phi (fun n => iter A f n bot) -> phi = f phi.
intros phi [Hub Hl].
apply Ha.
apply Hl; apply f_upper_bound_iterates; assumption.
assert (Hlubfphi : lub R (f phi) (fun n => f (iter A f n bot))).
apply (Hct (fun n => iter A f n bot) Hchain_iter_f_bot phi);split;assumption.
destruct Hlubfphi as [Hub' Hl'].
apply Hl'; apply upper_bound_iterates_add_f; assumption.
Qed.
     
Theorem lub_iterates_less_than_fixpoint :
  forall phi : A, lub R phi (fun n => iter A f n bot) ->
  forall psi : A, psi = f psi -> R phi psi.
intros phi [Hup Hl] psi Hfxp; apply Hl.
intros n; induction n.
simpl; auto.
simpl; rewrite Hfxp; apply f_monotonic; assumption.
Qed.

Theorem Tarski_least_fixpoint : (exists phi: A, least_fixpoint R f phi).
destruct (Hc (fun n => iter A f n bot) Hchain_iter_f_bot) as [phi Hlub].
exists phi.
split.
apply lub_iterates_fixpoint; auto.
apply lub_iterates_less_than_fixpoint; auto.
Qed.

Theorem least_fixpoint_lub_iterates :
  forall phi, least_fixpoint R f phi -> lub R phi (fun n => iter A f n bot).
intros phi Hlfp.
destruct (Hc (fun n => iter A f n bot) Hchain_iter_f_bot) as [psi Hlub'].
replace phi with psi.
trivial.
symmetry.
apply least_unique with (1:= Ha) (P:= fun x => x = f x) (2:= Hlfp).
split.
apply lub_iterates_fixpoint; auto.
apply lub_iterates_less_than_fixpoint; auto.
Qed.

End Tarski_theory.



(* SECTION: basic properties of the basic discrete order. *)

Set Implicit Arguments.

Inductive option_cpo (A:Set) : option A -> option A -> Prop :=
  option_cpo_none_bot : forall x: option A, option_cpo None x
| option_cpo_refl : forall x: option A, option_cpo x x.

Unset Implicit Arguments.

Hint Resolve option_cpo_refl option_cpo_none_bot.

Theorem antisym_option_cpo :
  forall A, antisymetric (option_cpo (A:=A)).
intros A x y H; inversion H.
intros H'; inversion H'; trivial.
trivial.
Qed.

Theorem transitive_option_cpo_aux :
  forall (A:Set) (x y: option A),
   option_cpo x y -> forall z, option_cpo y z -> option_cpo x z.
intros A x y H; inversion H.
intros; apply option_cpo_none_bot.
intros; trivial.
Qed.

Theorem transitive_option_cpo :
  forall (A:Set) (x y z: option A),
   option_cpo x y -> option_cpo y z -> option_cpo x z.
intros A x y z H1 H2 ; apply transitive_option_cpo_aux with y ; assumption.
Qed.

Theorem increasing_function_none_pred :
  forall A:Set, forall f : nat -> option A,
  (forall n, option_cpo (f n) (f (S n))) ->
  forall n, f n = None -> forall p, p <= n -> f p = None.
intros A f Hincr n Hfn p Hpn; generalize Hfn; clear Hfn.
elim Hpn.
auto.
intros m Hpm Hfmfp Hfsm; apply Hfmfp.
assert (Hcpo : option_cpo (f m) (f (S m))).
apply Hincr.
inversion Hcpo; auto.
rewrite <- Hfsm; auto.
Qed.

Theorem complete_option_cpo :
  forall (A:Set), complete (option_cpo (A:=A)).
intros A f Hincr.
elim (classic (forall n, f n = None)).
intros Hn; exists (None (A:=A)).
split.
intros p; rewrite Hn; apply option_cpo_none_bot.
intros; apply option_cpo_none_bot.
intros Hnotforall.
elim not_all_ex_not with (P:= fun x => f x = None) (1:= Hnotforall).
intros x Hneq; exists (f x).
split.
intros y.
case (le_gt_dec x y).
intros Hxley.
elim Hxley.
apply option_cpo_refl.
intros y' Hxley' Hcpo; generalize (Hincr y'); intros Hcpoy'; inversion Hcpoy'.
elim Hneq.
apply increasing_function_none_pred with y'.
assumption.
auto.
assumption.
assumption.
intros Hgt; assert (Hylex : y <= x).
omega.
generalize Hneq.
elim Hylex.
intros; apply option_cpo_refl.
intros x' Hylex' Hrec Hnotnone.
assert (Hcpo: option_cpo (f x')(f (S x'))).
apply Hincr.
inversion Hcpo.
assert (Hy: f y = None).
apply increasing_function_none_pred with x'; auto.
rewrite Hy; apply option_cpo_none_bot.
apply Hrec; rewrite H1; auto.
intros y Hup; apply Hup.
Qed.

(* SECTION: providing usable mathematica tools for a classical setting. *)

(* I add this axiom after discussions with Benjamin Werner and Hugo Herbelin *)

Axiom
  dependent_description' :
    forall (A:Type) (B:A -> Type) (R:forall x:A, B x -> Prop),
      (forall x:A,
          exists y : B x, R x y /\ (forall y':B x, R x y' -> y = y')) ->
       sigT (fun f : forall x:A, B x => (forall x:A, R x (f x))).

(* Extensionality is usually judged to be a tame axiom (in a classical
   setting.) *)

Axiom extensionality :
  forall A : Type, forall B : A -> Type,
    forall f g: forall x:A, B x,
      (forall x, f x = g x) -> f = g.

(* Once with the dependent description as above, we can transform
   any proof of single existence into a value. *)

Definition shift_ex_to_sigT :
  forall (A : Set)(P: A -> Prop),
    (exists y:A, P y /\ forall y':A, P y' -> y = y') ->
  sigT (fun y :A => P y).
intros A P Hex.
assert (X:sigT (fun f : forall x:unit, (fun x => A) x =>
                forall z:unit, 
                   (fun x:unit => fun y : (fun x => A) x => P y) 
                   z (f z))).
apply dependent_description' with (B:= (fun x:unit => A))
    (R := fun (x:unit) (y:A)  => P y).
auto.
destruct X as [f Hf].
exists (f tt); auto.
Qed.


(* SECTION: lifting an order to the function space. *)

Definition lift_order :=
   fun (A B:Set)(R:B->B->Prop)(f g:A->B)=>
     forall x, R (f x)(g x).

Theorem lift_reflexive :
  forall (A B:Set)(R:B->B->Prop),
  (forall x, R x x) -> forall f, (lift_order A B R) f f.
intros A B R Hr f x; apply Hr.
Qed.

Theorem lift_antisymetric :
  forall (A B:Set)(R:B->B->Prop),
     antisymetric R -> antisymetric (lift_order A B R).
intros A B R Ha f g HfRg HgRf.
apply extensionality with (B:= fun x:A => B).
intros x; apply (Ha (f x) (g x)).
apply (HfRg x).
apply (HgRf x).
Qed.

Theorem lift_transitive :
  forall (A B:Set)(R:B->B->Prop),
  (forall x y z, R x y -> R y z -> R x z) ->
  forall f g h, lift_order A B R f g ->
      lift_order A B R g h ->
      lift_order A B R f h.
intros A B R Ht f g h Hfg Hgh x.
apply Ht with (g x).
apply (Hfg x).
apply (Hgh x).
Qed.

Theorem lift_complete :
  forall (A B:Set)(R:B->B->Prop),
  antisymetric R ->
  complete R -> complete (lift_order A B R).
intros A B R Ha Hc f Hchain.
assert (Witness : sigT (fun g => forall x:A, lub R (g x) (fun n => f n x))).
apply dependent_description' with (A:=A)(B:=fun x:A => B)
  (R:= fun (z:A)(t:B) => lub R t (fun n => f n z)).
intros x.
destruct (Hc (fun n => f n x)) as [y Hlub].
intros n; apply (Hchain n).
exists y; split.
assumption.
intros y' Hluby'; apply unique_lub with (R:=R) (f:= fun n=> f n x); auto.
destruct Witness as [g Testimony].
exists g; unfold lift_order.
split.
intros n x; destruct (Testimony x) as [Hup _].
exact (Hup n).
intros g' Hupg' x.
destruct (Testimony x) as [_ Hl].
apply Hl.
intros n; apply (Hupg' n x).
Qed.

Theorem lub_lift_reverse :
  forall (A B:Set)(R:B->B->Prop)(c:nat->A->B)(l:A->B),
  lub (lift_order A B R) l c ->
  forall x, lub R (l x)(fun n => c n x).
intros A B R c l [Hub Hl] x.
split.
intros n; apply (Hub n x).
intros b' Hup'.
destruct (choice (fun z t => (z = x -> t = b')/\ (z<>x -> t = l z))) as
  [g' Hdef_g'].
intros z; case (classic (z = x)).
intros; exists b'; split.
auto.
intros; contradiction.
intros; exists (l z); auto.
split.
intros; contradiction.
auto.
assert (Hupg' : upper_bound (lift_order A B R) g' c).
intros n z; case (classic (z=x)).
intros Heqzx; assert(Hg'zeqb': g' z = b').
elim (Hdef_g' z); auto.
rewrite Hg'zeqb'.
rewrite Heqzx.
exact (Hup' n).
intros Hzneqx; assert(Hg'zeqlz: g' z= l z).
elim (Hdef_g' z); auto.
rewrite Hg'zeqlz.
exact (Hub n z).
assert (Hg'xeqb' : g' x = b').
elim (Hdef_g' x); auto.
rewrite <- Hg'xeqb'.
exact (Hl g' Hupg' x).
Qed.

(* The specific form of the Tarski least fixpoint theorem when used
  solely to define recursive functions. *)
Definition Tarski_fix' :
   forall (A B:Set)(f:(A->option B)->A->option B)
   (Hct : continuous (lift_order A (option B) (@option_cpo B))
              (lift_order A (option B) (@option_cpo B)) f), 
   A->option B.
intros A B f Hct.
assert (Hex := Tarski_least_fixpoint _ 
              (lift_order A (option B) (@option_cpo B))
              (fun f x => option_cpo_refl (f x))
              (lift_antisymetric A (option B)
                  (@option_cpo B)(antisym_option_cpo B))
              (lift_complete A _ _
                 (antisym_option_cpo B)(complete_option_cpo B))
              (fun x:A=>None)
              (fun g x => option_cpo_none_bot (g x)) f Hct).
assert (X:sigT (fun phi : A -> option B => 
                least_fixpoint (lift_order _ _ (@option_cpo B)) f phi)).
apply shift_ex_to_sigT.
destruct Hex as [the_v Hlfp].
exists the_v; split.
auto.
intros v' Hlfp'; unfold least_fixpoint in Hlfp, Hlfp';
 apply least_unique with (P:= fun x => x = f x)
    (1:=lift_antisymetric A _ _ (antisym_option_cpo B))(2:=Hlfp)(3:=Hlfp').
destruct X as [it _]; exact it.
Defined.

Axiom Tarski_fix :
   forall (A B:Set)(f:(A->option B)->A->option B)
   (Hct : continuous (lift_order A (option B) (@option_cpo B))
              (lift_order A (option B) (@option_cpo B)) f), 
   A->option B.

Implicit Arguments Tarski_fix' [A B].

Implicit Arguments Tarski_fix [A B].

Definition f_order (A B:Set) :=
  lift_order A (option B)(@option_cpo B).

Theorem f_order_reflexive :
  forall A B f, f_order A B f f.
intros A B f x; auto.
Qed.

Theorem Tarski'_fix_prop :
    forall (A B:Set) (f : (A -> option B)->A -> option B)
      (Hct: continuous (f_order A B)(f_order A B) f),
       least_fixpoint 
         (f_order A B) f (Tarski_fix' f Hct).
intros A B f Hct; unfold Tarski_fix'.
match goal with |- context[(shift_ex_to_sigT ?a ?b ?c)] =>
    case (shift_ex_to_sigT a b c)
end.
intros x H; exact H.
Qed.

Axiom Tarski_fix_prop :
    forall (A B:Set) (f : (A -> option B)->A -> option B)
      (Hct: continuous (f_order A B)(f_order A B) f),
       least_fixpoint 
         (f_order A B) f (Tarski_fix f Hct).

(* Properties of chains. *)

Theorem continuous_chain :
  forall (A B:Set)(R:A->A->Prop)
     R' (f:A->B), (forall x, R x x) -> continuous R R' f ->
    forall c, (forall n, R (c n)(c (S n)))->
              forall n, R' (f (c n))(f (c (S n))).
intros A B R R' f Hr Hct c Hch n;
    apply (continuous_imp_monotonic _ _ R _ _ Hr Hct (c n) (c (S n))).
auto.
Qed.

Theorem option_chain_lub_none :
  forall (A:Set)(c:nat -> option A),
  lub (@option_cpo A) None c -> forall n, c n = None.
intros A c [Hup _] n.
assert (Hcmp : option_cpo (c n) None).
apply (Hup n).
inversion Hcmp; trivial.
Qed.

Theorem option_chain_some :
 forall (A:Set)(c:nat -> option A)(n:nat)(v:A), 
   (forall n, option_cpo (c n)(c (S n))) ->
   c n = Some v -> lub (@option_cpo A) (Some v) c.
intros A c n v Hch Heq.
destruct (complete_option_cpo A c Hch) as [v' Hlub].
elim Hlub; intros Hup _; assert (Hup_n := Hup n); 
rewrite Heq in Hup_n; inversion Hup_n.
subst v'; auto.
Qed.

Theorem option_chain_some_ge :
  forall (A:Set)(c:nat->option A),
  (forall n, option_cpo (c n) (c (S n))) ->
  forall n v, c n = Some v -> forall m, (n <= m)%nat ->
    c m = Some v.
intros A c Hch n v Heq m Hle; induction Hle.
auto.
assert (Hch':= Hch m).
rewrite IHHle in Hch'; inversion Hch'; auto.
Qed.

Theorem chain_lub_some_witness1 :
   forall (A:Set) (c:nat -> option A)(v:A),
   lub (@option_cpo A) (Some v) c -> 
   (exists n:nat, c n = Some v).
intros A c v [Hup Hl].
elim (classic (forall n, c n = None)).
intros Hnone.
assert (Hnone_aux : option_cpo (Some v) None).
apply Hl.
intros n.
rewrite Hnone; auto.
inversion Hnone_aux.	
intros Hnot_all.
assert (Hex_not : exists n, c n <> None).
apply not_all_ex_not with (U:=nat)
  (P:= fun n => c n = None)(1:= Hnot_all).
destruct Hex_not as [n Hneq].
assert (Hopt: option_cpo (c n) (Some v)).
apply (Hup n).
exists n.
inversion Hopt; auto.
elim Hneq; auto.
Qed.

Theorem chain_lub_some_witness :
   forall (A B:Set) chain (a:A)(b:B)(f:A->option B), f a = Some b -> 
   lub (lift_order A (option B) (option_cpo (A:=B))) f chain  ->
   (exists n:nat, chain n a = Some b).
intros A B chain a b f Heq Hlub.
apply (chain_lub_some_witness1 B (fun n => chain n a) b).
generalize (lub_lift_reverse A (option B)(@option_cpo B) chain f Hlub a).
rewrite Heq; auto.
Qed.

Theorem Tarski_fix_iterates_witness :
    forall (A B:Set) (f : (A -> option B)->A -> option B)
      (Hct: continuous (f_order A B)(f_order A B) f)(x:A)(v:B),
    Tarski_fix f Hct x = Some v ->
    exists n, iter (A->option B) f n (fun a => None) x = Some v.
intros A B f Hct x.
assert (H:lub (f_order A B) (Tarski_fix f Hct)
              (fun n => iter _ f n (fun a=>None))).
apply least_fixpoint_lub_iterates.
unfold f_order; apply lift_reflexive; auto.
unfold f_order; apply lift_antisymetric.
apply antisym_option_cpo.
unfold f_order; apply lift_complete.
apply antisym_option_cpo.
apply complete_option_cpo.
intros g y; auto.
assumption.
apply Tarski_fix_prop.
intros v Heq.
apply chain_lub_some_witness with (A:=A)(B:=B)(1:= Heq)(2:=H).
Qed.

Theorem iterates_none_imp_fix_none :
  forall A B f (Hct : continuous (f_order A B)(f_order A B) f) x,
  (forall n, iter _ f n (fun z => None) x = None) ->
  Tarski_fix f Hct x = None.
intros A B f Hct x Happr.
caseEq (Tarski_fix f Hct x).
intros v Heq;
destruct (Tarski_fix_iterates_witness A B f Hct x v Heq) as
  [n Heq'].
rewrite (Happr n) in Heq'; discriminate.
trivial.
Qed.

(* SECTION: Defining the recursive function that is needed for the
   semantics of the langage of while loops and assignments. *)

(* SECTION: the order with which we work.  It is complete by construction. *)

Definition f_order' (A : Set) :=
  lift_order A (option A)(option_cpo (A:=A)).

Theorem f_order'_reflexive : forall (A:Set) g, f_order' A g g.
intros A g x; apply option_cpo_refl.
Qed.

Theorem f_order'_antisymetric : forall A:Set, antisymetric (f_order' A).
intros A; unfold f_order'; apply lift_antisymetric.
apply antisym_option_cpo.
Qed.

Theorem f_order'_complete : forall A:Set, complete (f_order' A).
intros A; unfold f_order'; apply lift_complete.
apply antisym_option_cpo.
apply complete_option_cpo.
Qed.

(* SECTION: the operation of composing on the left with an arbitrary
   function. *)

Definition comp_right (A:Set)(f g:A->option A)(x:A) :=
  match f x with None => None | Some r => g r end.

Implicit Arguments comp_right.

Theorem comp_right_continuous :
  forall (A:Set)(f:A->option A),
     continuous (f_order' A)(f_order' A)(comp_right f).
intros A f c Hc l [Hu Hl].
split.
intros n x; unfold comp_right; case (f x).
exact (Hu n).
apply option_cpo_none_bot.
intros g' Hu' x.
assert (Hc' : forall n, option_cpo (comp_right f (c n) x)
                                   (comp_right f (c (S n)) x)).
intros n; unfold comp_right; case (f x).
intros a; exact (Hc n a).
apply option_cpo_none_bot.
caseEq (f x).
intros a Hfxeqa.
unfold comp_right; rewrite Hfxeqa.
generalize (lub_lift_reverse A (option A) (option_cpo (A:=A)) c l 
            (conj Hu Hl) a).
intros [_ Hlla].
apply Hlla.
intros n; generalize (Hu' n x); unfold comp_right; rewrite Hfxeqa; trivial.
intros Hfxeqnone; unfold comp_right; rewrite Hfxeqnone.
apply option_cpo_none_bot.
Qed.

(* SECTION: the cond statement. *)

Definition cond (A B:Set)(t:A->option bool)(f g : A -> option B)(x:A) :=
   match t x with
     None => None
   | Some true => f x
   | Some false => g x
   end.

Implicit Arguments cond.


Theorem cond_continuous :
  forall A t F G,
    continuous (f_order' A)(f_order' A) F ->
    continuous (f_order' A)(f_order' A) G ->
    continuous (f_order' A)(f_order' A) (fun f => cond t (F f) (G f)).
intros A t F G Hcf Hcg c Hch l Hlub; split.
intros n x; unfold cond.
caseEq (t x).
intros [|] Heq.
destruct (Hcf c Hch l Hlub) as [Hupf Hlf].
apply (Hupf n x).
destruct (Hcg c Hch l Hlub) as [Hupg Hlg].
apply (Hupg n x).
intros; apply option_cpo_none_bot.
intros h Huph x; unfold cond.
caseEq (t x).
intros [ | ] Heqtx.
assert (Hlubf : lub (f_order' A)(F l)(fun n => F (c n))).
exact (Hcf c Hch l Hlub).
destruct (lub_lift_reverse _ _ _ _ _ Hlubf x) as [Hupx Hlx].
apply Hlx.
intros n.
generalize (Huph n x); unfold cond; rewrite Heqtx; trivial.

assert (Hlubg : lub (f_order' A)(G l)(fun n => G (c n))).
exact (Hcg c Hch l Hlub).
destruct (lub_lift_reverse _ _ _ _ _ Hlubg x) as [Hupx Hlx].
apply Hlx.
intros n.
generalize (Huph n x); unfold cond; rewrite Heqtx; trivial.

intros; apply option_cpo_none_bot.
Qed.

(* SECTION : continuity of constant functions. *)

Theorem lub_cst :
  forall (A:Set)(R:A->A->Prop)(a:A),(forall x, R x x)->lub R a (fun n=>a).
intros A R a Hrefl.
split.
intros n; apply Hrefl.
intros v Hup.
apply (Hup 0).
Qed.

Theorem continuous_cst :
  forall (A B:Set)(R:A->A->Prop)(R':B->B->Prop)(b:B),
  (forall x, R' x x) -> continuous R R' (fun x:A => b).
intros A B R R' b Hrefl c Hch l Hlub.
apply lub_cst; auto.
Qed.

Theorem exists_phi :
   forall (A : Set) (t : A -> option bool) (f : A -> option A),
       exists phi : A -> option A,
         least_fixpoint (f_order' A) 
           (fun g : A -> option A => 
               cond t (comp_right f g) (fun x : A => Some x))
           phi.
intros A t f.
assert (Hrefl : forall x, f_order' A x x).
unfold f_order'.
intros x; apply (lift_reflexive A (option A) (option_cpo(A:=A))).
apply option_cpo_refl.
assert (Ha : antisymetric (f_order' A)).
apply f_order'_antisymetric.

apply Tarski_least_fixpoint with 
   (R := f_order' A)
   (bot := fun x:A => None (A:=A))
   (f := fun g => cond t (comp_right f g)(fun x => Some x)).
assumption.
assumption.
unfold f_order'; apply lift_complete.
apply antisym_option_cpo.
apply complete_option_cpo.
intros g x; apply option_cpo_none_bot.
apply cond_continuous with (F:= comp_right f)
        (G:= fun (g:A->option A)(x:A) => Some x).
apply comp_right_continuous.
apply continuous_cst.
assumption.
Qed.

Definition phi (A:Set)(t:A->option bool)(f:A->option A) (* : A->option A *) :=
  Tarski_fix (fun g => cond t (comp_right f g)(fun y => Some y))
     (cond_continuous A t (fun g => comp_right f g)(fun g => (fun x => Some x))
        (comp_right_continuous A f)
         (continuous_cst _ _ (f_order' A)(f_order' A)(fun x=> Some x)
            (f_order'_reflexive A))).

(* Definition phi (A:Set)(t:A->option bool)(f:A->option A) :
  sigT (fun v : A->option A => 
          least_fixpoint (f_order' A)
            (fun x => cond t (comp_right f x)(fun x => Some x))
            v).
intros A t f.
apply shift_ex_to_sigT.
destruct (exists_phi A t f) as [phi Hlfp].
exists phi; split; auto.
intros y' Hlfp'; apply least_unique with 
   (P := fun x => x = (fun g => cond t (comp_right f g)(fun a=>Some a)) x)
   (x:= phi)(y:=y')
   (2:= Hlfp)(3:=Hlfp').
apply f_order'_antisymetric.
Qed.
*)

(* SECTION : Specializing the whole for the while language. *)

Open Scope Z_scope.

Inductive expr : Set :=
  var : Z -> expr
| num : Z -> expr
| plus : expr -> expr -> expr.

Inductive bexpr : Set :=
  greater : expr -> expr -> bexpr.

Inductive instr : Set :=
  assign : Z -> expr -> instr
| sequence : instr -> instr -> instr
| while : bexpr -> instr -> instr.

Definition state := list (Z*Z).

Fixpoint flookup (l:state)(x:Z) {struct l} : option Z :=
  match l with
    nil => None
  | (x',y)::tl =>
    match x' ?=  x with Eq => Some y | _ => flookup tl x  end
  end.

Fixpoint feval (l:state)(e : expr) {struct e} : option Z :=
  match e with
    var x => flookup l x
  | num n => Some n
  | plus e1 e2 => 
    match feval l e1 with
      None => None
    | Some v1 =>
       match feval l e2 with
        None => None
      | Some v2 => Some (v1+v2)
      end
    end
  end.

Definition evalb (l:state)(b:bexpr) : option bool :=
  match b with
    greater e1 e2 =>
    match feval l e1 with
      None => None
    | Some v1 =>
      match feval l e2 with
        None => None
      | Some v2 => 
        match v1 ?= v2 with
          Gt => Some true
        | _ => Some false
        end
      end
    end
  end.

Fixpoint fupdate (l:state)(x v : Z) {struct l} : option state :=
  match l with
    nil => None
  | (x', v')::tl => 
    match x' ?= x with
      Eq => Some ((x, v)::tl)
    | _ =>
      match fupdate tl x v with
        None => None
      | Some tl' => Some ((x', v')::tl')
      end
    end
  end.

Fixpoint ds (l:state) (i:instr) {struct i} : option state :=
  match i with
    assign x e =>
      match feval l e with
        None => None
      | Some v => fupdate l x v
      end
  | sequence i1 i2 =>
    match ds l i1 with 
      None => None
    | Some l' => ds l' i2
    end
  | while e i =>
    phi state (fun l' => evalb l' e)(fun l' => ds l' i) l
  end.

Inductive lookup : list (Z*Z) -> Z -> Z -> Prop :=
  lookup_found : forall x y tl, lookup ((x,y)::tl) x y
| lookup_rec : forall x x' y y' tl, x <> x' -> lookup tl x y -> 
                   lookup ((x',y')::tl) x y.

Inductive eval : list (Z*Z) -> expr -> Z -> Prop :=
  eval_var : forall l x v, lookup l x v -> eval l (var x) v
| eval_num : forall l n, eval l (num n) n
| eval_plus : 
    forall l e1 e2 n1 n2,
      eval l e1 n1 -> eval l e2 n2 -> eval l (plus e1 e2) (n1+n2).

Inductive update : list (Z*Z) -> Z -> Z -> list (Z*Z) -> Prop :=
  update_found : forall x y y' tl,
                   update ((x,y)::tl) x y' ((x,y')::tl)
| update_rec : forall x x' y y' tl tl', x<>x' ->
     update tl x y tl' -> update ((x',y')::tl) x y ((x',y')::tl').

Inductive b_eval : list (Z*Z) -> bexpr -> bool -> Prop :=
| be_greater1 : forall r e1 e2 v1 v2,
            eval r e1 v1 -> eval r e2 v2 ->
            v2 < v1 -> b_eval r (greater e1 e2) true
| be_greater2 : forall r e1 e2 v1 v2,
            eval r e1 v1 -> eval r e2 v2 ->
            v1 <= v2 -> b_eval r (greater e1 e2) false.

Inductive nat_sem : list (Z*Z) -> instr -> list (Z*Z) -> Prop :=
  nat_sem_assign : forall x e n l l',
     eval l e n -> update l x n l' -> nat_sem l (assign x e) l'
| nat_sem_sequence : forall i1 i2 l l1 l2,
   nat_sem l i1 l1 -> nat_sem l1 i2 l2 -> 
   nat_sem l (sequence i1 i2) l2
| nat_sem_while_true : forall e i l l1 l2,
    b_eval l e true -> nat_sem l i l1 ->
    nat_sem l1 (while e i) l2 ->
    nat_sem l (while e i) l2
| nat_sem_while_false : forall e i l,
    b_eval l e false ->
    nat_sem l (while e i) l.

Theorem flookup_sound :
  forall l x v, flookup l x = Some v -> lookup l x v.
induction l as [ | [x' v'] l IHl].
simpl; intros; discriminate.
simpl; intros x v; 
  generalize (Zcompare_elim (x' = x) (x' <> x) (x' <> x) x' x).
case (x'?=x).
intro He.
assert (He': x'=x).
apply He; intros; omega.
intros Hv; injection Hv; intros Hv1; subst v' x'.
constructor.

intros He.
assert (He' : x'<>x).
apply He; intros; omega.
intros; apply lookup_rec; auto.

intros He.
assert (He' : x'<>x).
apply He; intros; omega.
intros; apply lookup_rec; auto.
Qed.

Theorem feval_sound :
  forall l e v, feval l e = Some v -> eval l e v.
intros l e; induction e; intros v.
intros; apply eval_var.
apply flookup_sound; auto.
simpl; intros Hf; injection Hf; intros; subst v; apply eval_num.
simpl; generalize IHe1; case (feval l e1).
intros v1 IHe1'; generalize IHe2; case (feval l e2).
intros v2 IHe2' Hres; injection Hres; intros Hres'; subst v.
apply eval_plus; auto.
intros; discriminate.
intros; discriminate.
Qed.

Theorem beval_sound :
  forall l b v, evalb l b = Some v -> b_eval l b v.
intros l [e1 e2] v; simpl.
generalize (feval_sound l e1); case (feval l e1).
intros v1 He1; generalize (feval_sound l e2); case (feval l e2).
intros v2 He2; generalize (Zcompare_elim (v1 = v2)(v1 < v2)(v1 > v2) v1 v2).
case (v1?=v2); intros Hc Hres; injection Hres; intros; subst v.
apply be_greater2 with v1 v2; auto.
omega.
apply be_greater2 with v1 v2; auto.
omega.
apply be_greater1 with v1 v2; auto.
omega.
intros; discriminate.
intros; discriminate.
Qed.

Theorem fupdate_sound :
  forall l x v l', fupdate l x v = Some l' -> update l x v l'.
induction l as [ | [x' v'] tl IHtl]; intros x v l'.
simpl; intros; discriminate.
simpl; generalize (Zcompare_elim (x' = x)(x' <> x)(x' <> x) x' x).
case (x'?=x).
intros He; assert (x' = x).
omega.
intros Hres; injection Hres; intros; subst x' l'.
apply update_found.

intros He; assert (x' <> x).
omega.
generalize (IHtl x v); case (fupdate tl x v).
intros tl' Htl' Hres; injection Hres; intros; subst l'.
apply update_rec; auto.
intros; discriminate.

intros He; assert (x' <> x).
omega.
generalize (IHtl x v); case (fupdate tl x v).
intros tl' Htl' Hres; injection Hres; intros; subst l'.
apply update_rec; auto.
intros; discriminate.
Qed.

Theorem ds_sound :
  forall l i l', ds l i = Some l' -> nat_sem l i l'.
intros l i; generalize l; clear l.
induction i.

intros l l'; simpl.
generalize (feval_sound l e); case (feval l e).
intros v He; generalize (fupdate_sound l z v); case (fupdate l z v).
intros l'' Hu Hres; injection Hres; intros; subst l''.
apply nat_sem_assign with v; auto.
intros; discriminate.
intros; discriminate.

simpl; intros l l'; generalize (IHi1 l); case (ds l i1).
intros l1 Hi1; generalize (IHi2 l1); case (ds l1 i2).
intros l2 Hi2 Hres; injection Hres; intros; subst l'.
apply nat_sem_sequence with l1; auto.
intros; discriminate.
intros; discriminate.

simpl; intros l l'.
unfold phi.
match goal with |- context [cond_continuous ?a ?b ?c ?d ?e ?f] =>
   generalize (cond_continuous a b c d e f)
end.
intros c.
generalize 
    (Tarski_fix_prop state state 
        (fun g=>cond (fun l': state => evalb l' b)
              (comp_right (fun l' : state => ds l' i) g)
              (fun y : state => Some y)) c).
generalize (Tarski_fix (fun g=>cond (fun l': state => evalb l' b)
              (comp_right (fun l' : state => ds l' i) g)
              (fun y : state => Some y)) c).
clear c; intros phi0 Hphi0 Hres.
assert (Hlub : lub (f_order' state) phi0
                (fun n =>
                   iter _
                     (fun x => cond (fun l' => evalb l' b)
                                 (comp_right (fun l'=> ds l' i) x)
                                 (fun s => Some s)) n
                     (fun x : state => None))).
apply least_fixpoint_lub_iterates.
apply f_order'_reflexive.
apply f_order'_antisymetric.
apply f_order'_complete.
intros g x; apply option_cpo_none_bot.
apply cond_continuous with 
   (F:=comp_right (fun l' => ds l' i))
   (G:=fun (g:state->option state)(x:state)=>Some x).
apply comp_right_continuous.
apply continuous_cst.
apply f_order'_reflexive.
assumption.
cut (exists n:nat, iter (state->option state)
             (fun x : state -> option state =>
               cond (fun l' : state => evalb l' b)
                 (comp_right (fun l' : state => ds l' i) x)
                 (fun s : state => Some s)) n (fun _ : state => None) l =
                 Some l').

intros H; elim H; clear H; intros n; generalize l.
clear Hlub Hres Hphi0 l.
induction n; intros l.
simpl; intros; discriminate.
simpl; unfold cond at 1.  
generalize (beval_sound l b); case (evalb l b).
intros [|] Hevalb.
unfold comp_right at 1.
generalize (IHi l); case (ds l i).
intros l1 Hds_i Hds_w; apply nat_sem_while_true with l1; auto.
intros; discriminate.
intros Hres; injection Hres; intros; subst l'.
apply nat_sem_while_false; auto.
intros; discriminate.
apply (chain_lub_some_witness state state
          _ l l' phi0 Hres Hlub).
Qed.



